/*---------------------------------------------------------------

  PROGRAM : neighbor.c
  DATE    : 8/3/94
  AUTHOR  : Wasiuddin Wahid, wasi@media.mit.edu
  ______________________________________________________________

  PROGRAM DESCRIPTION HERE









  ---------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define OPTIONS "m:x:k:l:o:n:t:w:u:c:a:b:f"
#define MAX_CHARS 256

char *usage = "\t-m dir_known_faces -x dir_test_faces\n\t\
\t\t-k list_known_faces -l list_test_faces\n\t\
\t\t-o output_file -t threshold -w weights\n\t\
\t\t-u biases [-c conf] [-n top_n_matches] [-a lower coeff]\n\t\
\t\t[-b upper coeff] -f [read on fs of Probe]";

char *help = " 
Nearest Neighbor Pattern Matching\n
-m dir_known_faces  \t directory of  known faces \n\
-x dir_test_faces   \t directory of test faces \n\
-k list_known_faces \t list file of known faces \n\
-l list_test_faces  \t list file of test faces\n\
-o output_file      \t output file\n\
-n top_n_matches    \t print out top N matches only(default 10) \n\
-t threshold        \t found flag threshold\n\
-w weights          \t weights file\n\
-u biases           \t bias file\n\
-c conf             \t confidence factor (default: 1000)\n\
-a lower coeff      \t lower coefficient (default: first)\n\
-b upper coeff      \t upper coefficient (default: last)\n\
-f                  \t read only fs of Probe\n\n";

extern int optind;
extern char *optarg;

/*------- command line defaults ------- */
float conf_factor = 1e3;

/* ---------------------DEFINITIONS OF STRUCTURES------------------------*/
/*
  A tface(test_face) consists of the name of the file containing the test face
  and a pointer to the structure "match".
  A match is a structure containing the name of the file to which the test_face
  is similar, the distance from that face and a pointer to the next 
  matched face 
  
  A kface(known_face) on the other hand is a structure containing it's name
  and four float arrays for the four sets of coefficients f, q, h and p.
  
  A test is basically a structure containing four float arrays for the four 
  set of coefficients f, q, h and p. These are for storing the coefficients
  for each test_face as you read them in one at a time.
  -------------------------------------------------------------------------*/


typedef struct match *Matchptr;

typedef struct match {
  char *name;
  float distance;
  Matchptr next;
} Match;

typedef struct tface{
  char *name;
  Matchptr similar;
} Tface;

typedef struct kface{
  char *name;
  float *fcoeffs;
  float *qcoeffs;
  float *hcoeffs;
  float *pcoeffs;
} Kface;

typedef struct test{
  float *fcoeffs;
  float *qcoeffs;
  float *hcoeffs;
  float *pcoeffs;
} Test;

int number_files(char list[]);
Matchptr add_match(Matchptr root, Matchptr new, int n);
float find_distance(float *test, float *known, int no);
int number_points(char descriptor[]);
void read_gcoeffs(char dir[], char list[], int no_points, Kface *kfaces);
void read_pcoeffs(char dir[], char line[], Test artest, int no_points, int f_flag);
void read_weights(char weight_file[], float *weights);
void print_match(Tface *faces, int n,int no_files, char out[], float thres, char comline[]);
int print_file(Tface faces, int n, FILE *fout, float thres, int count);
void kill_list(Matchptr list);
int empty_list(Matchptr list);
/* char *strdup(char *s); */





/*---------------------------- MAIN ----------------------------------*/
main(int argc, char **argv)
{
  
  int no_files, no_kfiles, no_points, file_no, point, i, c, n=10, rec, a=1, b=0, slength, interval, k, j, count =0;
  char *progname, comline[MAX_CHARS], ktype, ttype;

  char known_dir[MAX_CHARS];
  char test_dir[MAX_CHARS];
  char klist[MAX_CHARS];
  char tlist[MAX_CHARS];
  char out_file[MAX_CHARS];
  char line[MAX_CHARS];
  char kfile[MAX_CHARS];
  char tfile[MAX_CHARS];
  char the_kfile[MAX_CHARS];
  char descriptor[MAX_CHARS];
  char weight_file[MAX_CHARS];
  char bias_file[MAX_CHARS];
  
  FILE *fknown, *ftest, *ftlist, *fklist, *fp, *fout;
  Tface tfaces;
  Kface *kfaces;
  Test artest;
  Matchptr a_match;
  float a_point, dist, thres, weights[16], weight, biases[16];
  

  /* These are to use the read_descriptor in Baback's files */
  int nframe, sets, bytes_pixel, ydim; 


  
  /* Required Input flags */
  int errflag     = 0;
  int inflag      = 0;
  int testflag    = 0;
  int kflag       = 0;
  int tflag       = 0;
  int outflag     = 0;
  int nflag       = 0;
  int thresflag   = 0;
  int weight_flag = 0;
  int bias_flag   = 0;
  int f_flag      = 0;
  
  progname = argv[0];
  for (i=0; i<argc; i++){
    strcat(comline, argv[i]), strcat(comline, " ");
  }
  while ( (c = getopt(argc, argv, OPTIONS)) != EOF)
    switch(c){
      
    case 'm':
      strcpy(known_dir, optarg);
      inflag = 1;
      break;
      
    case 'x':
      strcpy(test_dir, optarg);
      testflag = 1;
      break;
      
    case 'k':
      strcpy(klist, optarg);
      kflag = 1;
      break;
      
    case 'l':
      strcpy(tlist, optarg);
      tflag = 1;
      break;
      
    case 'o':
      strcpy(out_file, optarg);
      outflag = 1;
      break;
      
    case 'n':
      n = atoi(optarg);
      nflag = 1;
      break;
      
    case 't':
      thres = atof(optarg);
      thresflag = 1;
      break;
      
    case 'w':
      strcpy(weight_file, optarg);
      weight_flag = 1;
      break;
      
    case 'u':
      strcpy(bias_file, optarg);
      bias_flag = 1;
      break;
      
    case 'c':
      conf_factor = atof(optarg);
      break;
      
    case 'a':
      a = atoi(optarg);
      break;
      
    case 'b':
      b = atoi(optarg);
      break;
      
    case 'f':
      f_flag = 1;
      break;
      
    case '?':
      errflag = 1;
      break;
    } /* switch (c){ */
  
  /* Command line check */
  if (errflag || !inflag || !testflag || !thresflag || !kflag || !tflag || !outflag || !weight_flag || !bias_flag){
    fprintf(stderr, "\nUSAGE: %s %s\n %s\n\n", progname, usage, help);
    exit(1);
  }
 
  
  /* Calculate the no. of test and known faces and allocate memory for the 
     arrays to store them */
  
  no_files = number_files(tlist);
  no_kfiles = number_files(klist);

  /* Wasi's Mistake: kfaces = (Kface *) calloc(no_files, sizeof(Kface));  */
  /* ----- baback's fix -------- */
  kfaces = (Kface *) calloc(no_kfiles, sizeof(Kface));

  if (kfaces == NULL){
    fprintf(stderr, "\nERROR: Memory for kfaces cannot be allocated: \n\n");
    exit(3);
  }
  
  /* Calculate the total no. of coefficients */
  sprintf(descriptor, "%s/f", test_dir);
  read_descriptor(descriptor, &nframe, &sets, &bytes_pixel,&ydim, &no_points);

  /* Read in the weights and biases from the appropriate  files */
  read_weights(weight_file, weights);
  read_weights(bias_file, biases);
  
  /* Read in the no of coefficients for gallery */
  read_gcoeffs(known_dir, klist, no_points, kfaces); 
 
  if ( (ftlist = fopen(tlist, "r")) == NULL) {
    fprintf(stderr, "\nERROR: Input file %s cannot be opened\n\n", tlist);
    fprintf(stderr, "\nThis should be the list file for the Probe faces\n\n");
    exit(2);
  }
    

  /* the interval relevant in the distance calculation is */
  /* difference between a & b                             */
  if (!b){
    b = no_points;
  }
  interval = b - a + 1;
  a = a - 1;
   
  if (no_kfiles < n){
    n = no_kfiles;
  }
  
  if ( (fout = fopen(out_file, "w")) == NULL){
    fprintf(stderr, "\nERROR: output file %s cannot be opened\n", out_file);
    fprintf(stderr, "This is the output file to which results are to be written\n");
    exit(2);
  }
  fprintf(stderr, "Command line : %s \n\n", comline);
  fprintf(fout, "Command Line: %s\n\n", comline);
  fprintf(fout, "<Probe number> <Match order> <Gallery Match Number> <Found Flag> <Confidence>\n\n");
  
  file_no = 0;
  /* Loop over the input file list and calculate the nearest neigbors */
  while ( (fscanf(ftlist, "%s", line)) != EOF){
    if ( (strncmp(line, "#", 1) != 0) && strlen(line) > 1) {
      
      /* Read in coeffs in the appropriate positions */
      artest.fcoeffs = (float *) calloc(no_points, sizeof(float));
      artest.qcoeffs = (float *) calloc(no_points, sizeof(float));
      artest.hcoeffs = (float *) calloc(no_points, sizeof(float));
      artest.pcoeffs = (float *) calloc(no_points, sizeof(float));
      
      if ( (artest.fcoeffs == NULL) || (artest.qcoeffs == NULL) || (artest.hcoeffs == NULL) || (artest.pcoeffs == NULL)){
	fprintf(stderr, "\nERROR: memory cannot be allocated for artest");
	exit(3);
      }
      
      read_pcoeffs(test_dir, line, artest, no_points, f_flag);
            
      tfaces.name = strdup(line);
      tfaces.similar = NULL;
            
      fprintf(stdout, "Matching probe: %s\n", line);
      
      /* find out what type of face it is: f, q, h or p */
      slength = strlen(line);
      for (k=0; k<slength; k++){
	if ( (line[k] >= 'a') && (line[k] <= 'z') ){
	  ttype = line[k];
	  break;
	}
      }      
      /* loop through the array of Kfaces finding the nearest neighbor */
      for (i=0; i < no_kfiles; i++){
	
	/* find the type of known face */
	slength = strlen( kfaces[i].name);
	for (k=0; k<slength; k++){
	  if ( (kfaces[i].name[k] >= 'a') && (kfaces[i].name[k] <= 'z')){
	    ktype = kfaces[i].name[k];
	    break;
	  }
	}	
	
	/* find out which coefficients and weights to use */
	if (ttype == 'f'){
	  if (ktype == 'f'){
	    dist = find_distance(artest.fcoeffs+a,kfaces[i].fcoeffs+a, interval);
	    dist = 100 * log(conf_factor)/log(dist + conf_factor);
	    dist = weights[0] * dist + biases[0];
	  } 
	  else {
	    if (ktype == 'q'){
	      dist = find_distance(artest.qcoeffs+a,kfaces[i].fcoeffs+a, interval);
	      dist = 100 * log(conf_factor)/log(dist + conf_factor);
	      dist = weights[1] * dist + biases[1];
	    }
	    else {
	      if (ktype == 'h'){
		dist = find_distance(artest.hcoeffs+a,kfaces[i].fcoeffs+a, interval);
		dist = 100 * log(conf_factor)/log(dist + conf_factor);
		dist = weights[2] * dist + biases[2];
	      }
	      else {
		if (ktype == 'p'){
		  dist = find_distance(artest.pcoeffs+a,kfaces[i].fcoeffs+a, interval);
		  dist = 100 * log(conf_factor)/log(dist + conf_factor);
		  dist = weights[3] * dist + biases[3];
		  
		}
	      }
	    }
	  }
	}  /* if (ttype = 'f') */
	else {
	  if (ttype == 'q'){
	    if (ktype == 'f'){
	      dist = find_distance(artest.fcoeffs+a,kfaces[i].qcoeffs+a, interval);
	      dist = 100 * log(conf_factor)/log(dist + conf_factor);
	      dist = weights[4] * dist + biases[4];
	    }
	    else{
	      if (ktype == 'q'){
		dist = find_distance(artest.qcoeffs+a,kfaces[i].qcoeffs+a, interval);
		dist = 100 * log(conf_factor)/log(dist + conf_factor);
		dist = weights[5] * dist + biases[5];
	      }
	      else{
		if (ktype == 'h'){
		  dist = find_distance(artest.hcoeffs+a,kfaces[i].qcoeffs+a, interval);
		  dist = 100 * log(conf_factor)/log(dist + conf_factor);
		  dist = weights[6] * dist + biases[6];
		}
		else{
		  dist = find_distance(artest.pcoeffs+a, kfaces[i].qcoeffs, interval);
		  dist = 100 * log(conf_factor)/log(dist + conf_factor);
		  dist = weights[7] * dist + biases[7];
		}
	      }
	    }
	  }  /* if (ttype == q)... */
	  else{
	    if ( ttype == 'h'){
	      if (ktype == 'f'){
		dist = find_distance(artest.fcoeffs+a, kfaces[i].hcoeffs+a, interval);
		dist = 100 * log(conf_factor)/log(dist + conf_factor);
		dist = weights[8] * dist + biases[8];
	      }
	      else{
		if (ktype == 'q'){
		  dist = find_distance(artest.qcoeffs+a, kfaces[i].hcoeffs, interval);
		  dist = 100 * log(conf_factor)/log(dist + conf_factor);
		  dist = weights[9] * dist + biases[9];
		}
		else {
		  if (ktype == 'h'){
		    dist = find_distance(artest.hcoeffs+a, kfaces[i].hcoeffs+a, interval);
		    dist = 100 * log(conf_factor)/log(dist + conf_factor);
		    dist = weights[10] * dist + biases[10];
		  }
		  else{
		    dist = find_distance(artest.pcoeffs+a, kfaces[i].hcoeffs+a, interval);
		    dist = 100 * log(conf_factor)/log(dist + conf_factor);
		    dist = weights[11] * dist + biases[11];
		  }
		}
	      }
	    }/* if (ttype == 'h').... */
	    else {
	      if (ktype == 'f'){
		dist = find_distance(artest.fcoeffs+a, kfaces[i].pcoeffs+a, interval);
		dist = 100 * log(conf_factor)/log(dist + conf_factor);
		dist = weights[12] * dist + biases[12];
	      }
	      else{
		if (ktype == 'q'){
		  dist = find_distance(artest.qcoeffs+a, kfaces[i].pcoeffs+a,interval);
		  dist = 100 * log(conf_factor)/log(dist + conf_factor);
		  dist = weights[13] * dist + biases[13];
		}
		else{
		  if (ktype == 'h'){
		    dist = find_distance(artest.hcoeffs+a, kfaces[i].pcoeffs+a, interval);
		    dist = 100 * log(conf_factor)/log(dist + conf_factor);
		    dist = weights[14] * dist + biases[14];
		  }
		  else{
		    dist = find_distance(artest.pcoeffs+a, kfaces[i].pcoeffs+a, interval);
		    dist = 100 * log(conf_factor)/log(dist + conf_factor);
		    dist = weights[15] * dist + biases[15];
		  }
		}
	      }
	    } /* third else....*/
	  } /* Second else ...*/
	}    /*  first else... */
	
	/* Create a match and add to the list of matches for that face */
	/* Allocate memory for the struct a_match of type Match for later use */
	a_match = (Matchptr) calloc(1, sizeof(Match));
	if (a_match == NULL){
	  fprintf(stderr, "\nERROR: Memory cannot be allocated for a_match\n\n");
	  exit(3);
	}
	
	a_match->name = strdup(kfaces[i].name);
	a_match->distance = dist;
	a_match->next = NULL;
	
	tfaces.similar = add_match(tfaces.similar, a_match, n);

      } /* for (i= 0; i<no_kfiles; ....*/
      
      file_no++;
      fprintf(fout, "\n");
      count =  print_file(tfaces, n, fout, thres, count);
      fprintf(fout, "\n");
      fflush(fout);
      kill_list(tfaces.similar);

    } /* if ( (strncmp........   */
  }   /* while ( (fscanf(.....   */
  
  fclose(ftlist);
  fprintf(stdout, "%f\n", (float) count/no_files);
  fprintf(stderr, "\n\nThe recognition rate is  %f  (%d/%d)\n", 
	  (float) count/no_files, count, no_files);
  fprintf(fout, "\n\nThe recognition rate is  %f  (%d/%d)\n", 
	  (float) count/no_files, count, no_files);
  
  fclose(fout);
  
} /* main */



void read_pcoeffs(char dir[],char line[], Test artest, int no_points, int f_flag)
{
  /* Requires  : The directory, dir contain four sub-directories f, q, h and p
                 containing coefficients and having the files line. Line 
		 should be a file in the standard format.
     effects   : Reads in the coefficients of the face line from each of 
                 the sub-directories f, q, h and p and store them in the 
		 structure artest. 
     modifies  : The structure artest                                       */

  int point = 0;
  float a_point;
  FILE *ftest;
  char tfile[MAX_CHARS];
    
  /* Allocate the necessary memory space for artest which is just 
     a structure Test which will be used for storing the coeffs for
     each test file */
  
  
  sprintf(tfile, "%s/f/%s", dir, line);
  if ( (ftest = fopen(tfile, "r")) == NULL){
    fprintf(stderr, "\nERROR: Input file %s cannot be opened in the function read_pcoeffs\n\n", tfile);
    fprintf(stderr, "This is a probe file\n");
    exit(2);
  }
  while ( (fscanf(ftest, "%f", &a_point)) != EOF){
    artest.fcoeffs[point] = a_point;
    point++;
  }
  fclose(ftest);
  
  if (!f_flag){
    point = 0;
    sprintf(tfile, "%s/q/%s", dir, line);
    if ( (ftest = fopen(tfile, "r")) == NULL){
      fprintf(stderr, "\nERROR: Input file %s cannot be opened in the function read_pcoeffs\n\n", tfile);
      fprintf(stderr, "This is a probe file\n");
      exit(2);
    }
    while ( (fscanf(ftest, "%f", &a_point)) != EOF){
      artest.qcoeffs[point] = a_point;
      point++;
    }
    fclose(ftest);
    
    point = 0;
    sprintf(tfile, "%s/h/%s", dir, line);
    if ( (ftest = fopen(tfile, "r")) == NULL){
      fprintf(stderr, "\nERROR: Input file %s cannot be opened in the function read_pcoeffs\n\n", tfile);
      fprintf(stderr, "This is a probe file\n");
      exit(2);
    }
    while ( (fscanf(ftest, "%f", &a_point)) != EOF){
      artest.hcoeffs[point] = a_point;
      point++;
    }
    fclose(ftest);
    
    point = 0;
    sprintf(tfile, "%s/p/%s", dir, line);
    if ( (ftest = fopen(tfile, "r")) == NULL){
      fprintf(stderr, "\nERROR: Input file %s cannot be opened in the function read_pcoeffs\n\n", tfile);
      fprintf(stderr, "This is a probe file\n");
      exit(2);
    }
    while ( (fscanf(ftest, "%f", &a_point)) != EOF){
      artest.pcoeffs[point] = a_point;
      point++;
    }
    fclose(ftest);
  }
} /* void read_pcoeffs..... */



int number_points(char descriptor[])
{
  /* requires : The descriptor be a file of the standard format
     effects  : extracts the no. of points info in the file and returns
     modifies : nothing                                         */


  int i;
  char line[MAX_CHARS], junk[MAX_CHARS];
  FILE *fp;

  if ( (fp=fopen(descriptor, "r")) == NULL) {
    fprintf(stderr, "\nERROR: Input file %s cannot be opened\n", descriptor);
    fprintf(stderr, "This is a descriptor file\n");
    exit(2);
  }

  fgets(line, MAX_CHARS, fp);
  fgets(line, MAX_CHARS, fp);
  fscanf(fp, "%s %i %s", line, &i, junk);
 
  fclose(fp);
  return(i);
}

  

int number_files(char list[])
{
  /* requires: list be a list of filenames of standard format
     effects : It counts and returns the number of filenames in list
     modifies: nothing                                                 */


  int no_files;
  char line[MAX_CHARS];
  FILE *flist;
  
  
  /* Loop over input list file and calculate total no. of 
     test files and create the array of tfaces */
  if ( (flist = fopen(list, "r")) == NULL){
    fprintf(stderr, "\nERROR: Input file %s cannot be opened \n\n", list);
    fprintf(stderr, "This is a list file and can be either probes or Gallery\n");
    exit(2);
  }
  no_files = 0;
  while (fscanf(flist, "%s", line) != EOF){
    if ( (strncmp(line, "#", 1) != 0) && (strlen(line) > 1))
      no_files++;
  }
  
  fclose(flist);
  return (no_files);
}



void read_gcoeffs(char dir[], char list[], int no_points, Kface *kfaces)
{

  /* requires: The dir consists of four directories called f, q, h and p. 
               These directories should have the files specified in list. 
	       All four files have the same name. Memory should have been 
	       allocated to the array kfaces.
     effects : The function reads in the coefficients from each directory in 
               the corresponding member in tface
     modifies: kfaces */

  int file_no = 0;
  int point = 0;
  FILE *fklist, *fknown;
  char line[MAX_CHARS], kfile[MAX_CHARS];
  float a_point;

  if ( (fklist = fopen(list, "r")) == NULL){
    fprintf(stderr, "\nERROR: the file %s cannot be opened in the function read_kcoeffs\n\n", list);
    exit(2);
    fprintf(stderr, "This is a list file for Gallery files\n");
  }

  /* Open the list file and read in all the coeffs */
  while (fscanf(fklist, "%s", line) != EOF){
    if ( (strncmp(line, "#", 1) != 0) || (strlen(line) >1)){
     
      /* Allocate the necessary memory space */
      /* ------- Baback's hack --------------------
      fprintf(stdout,"1: %s(%d)\n", line, file_no);
      fprintf(stdout,"2: %s\n", kfaces[file_no].name);
      ------------------------------------------------- */
      kfaces[file_no].name = strdup(line);
      kfaces[file_no].fcoeffs = (float *) calloc(no_points, sizeof(float));
      kfaces[file_no].qcoeffs = (float *) calloc(no_points, sizeof(float));
      kfaces[file_no].hcoeffs = (float *) calloc(no_points, sizeof(float));
      kfaces[file_no].pcoeffs = (float *) calloc(no_points, sizeof(float));
     
      if ( (kfaces[file_no].fcoeffs == NULL) || (kfaces[file_no].qcoeffs == NULL) || (kfaces[file_no].hcoeffs == NULL) || (kfaces[file_no].pcoeffs == NULL)){
	fprintf(stderr, "\nERROR: Memory cannot be allocated in the function read_kcoeffs\n\n");
	exit(3);
      }

      /* Read the f coeffs */
      sprintf(kfile, "%s/f/%s", dir, line);
      point = 0;
      if ( (fknown = fopen(kfile, "r")) == NULL){
	fprintf(stderr, "\nERROR: Input file %s cannot be opened\n\n", kfile);
	fprintf(stderr, "This is a known/Gallery file\n");
	exit(2);
      }
      while ( (fscanf(fknown, "%f", &a_point) != EOF)){
	kfaces[file_no].fcoeffs[point] = a_point;
	point++;
      }
      fclose(fknown);
      
      /* Read i all the q coeffs */
      sprintf(kfile, "%s/q/%s", dir, line);
      point = 0;
      if ( (fknown = fopen(kfile, "r")) == NULL){
	fprintf(stderr, "\nERROR: Input file %s cannot be opened\n\n", kfile);
	fprintf(stderr, "This is a known/Gallery file\n");
	exit(2);
      }
      while ( (fscanf(fknown, "%f", &a_point) != EOF)){
	kfaces[file_no].qcoeffs[point] = a_point;
	point++;
      }
      fclose(fknown);

      /*Read all the h coeffs */
      sprintf(kfile, "%s/h/%s", dir, line);
      point = 0;
      if ( (fknown = fopen(kfile, "r")) == NULL){
	fprintf(stderr, "\nERROR: Input file %s cannot be opened\n\n", kfile);
	fprintf(stderr, "This is a known/Gallery file\n");
	exit(2);
      }
      while ( (fscanf(fknown, "%f", &a_point) != EOF)){
	kfaces[file_no].hcoeffs[point] = a_point;
	point++;
      }
      fclose(fknown);
      
      /* Read all the p coeffs */
      sprintf(kfile, "%s/p/%s", dir, line);
      point = 0;
      if ( (fknown = fopen(kfile, "r")) == NULL){
	fprintf(stderr, "\nERROR: Input file %s cannot be opened\n\n", kfile);
	fprintf(stderr, "This is a known/Gallery file\n");
	exit(2);
      }
      while ( (fscanf(fknown, "%f", &a_point)) != EOF){
	kfaces[file_no].pcoeffs[point] = a_point;
	point++;
      }
      fclose(fknown);

      file_no++;
      
    } /* if ( (strncmp... */
  }   /* while (fscanf... */
} /* void read_kcoeffs */



void read_weights(char weight_file[], float *weights)
{
  /* Requires  : The weight file be of the standard format. A weight file 
                 consists of an ASCII matrix as follows:
		 1   2   3   4
		 5   6   7   8
		 9   10  11 12
		 13  14  15 16
     effects   : The function reads each entry into the array weights in the
                 same order as that shown above.
     modifies  : weights                                            */


  int i;
  FILE *fp;
  float a, b, c, d;

  if ( (fp = fopen(weight_file, "r")) == NULL) {
    fprintf(stderr, "\nERROR: Input file %s cannot be opened in the function read_weights\n\n", weight_file);
    fprintf(stderr, "This is either the weights or the biases  file\n");
    exit(2);
  }
  
  for (i=0; i<=12; i=i+4){
    fscanf(fp, "%f %f %f %f", &a, &b, &c, &d);
    weights[i] = a;
    weights[i+1]=b;
    weights[i+2]=c;
    weights[i+3]=d;
  }
  fclose(fp);
} /* void read_weights.... */



float find_distance(float *test, float *known, int no)
{
  /* Requires: no be less than the size of test or known
     effects : finds the Euclidean distance between the corresponding 
               coefficients stored in test and known
     modifies: nothing                                         */
  
  int i;
  float sum=0;
  
  for (i=0; i<no; i++){
    sum += ( (test[i] - known[i]) * (test[i]-known[i]));
  }
  return (sqrt(sum));
} /* float find_distance ....... */



void print_match(Tface *faces, int n, int no_files, char out[], float thres, char comline[])
{
  int i, j,k, count, l, rec=0;
  float d, e;
  Matchptr temp;
  FILE *fp;
  char test[MAX_CHARS], known[MAX_CHARS];
  
  if ( (fp = fopen(out, "w")) == NULL){
    fprintf(stderr, "\nERROR: Input file %s cannot be opened\n\n", out);
    exit(2);
  }
  
  fprintf(stderr, "Command line : %s \n\n", comline);
  fprintf(fp, "Command Line: %s\n\n", comline);
  fprintf(fp, "<Probe number> <Match order> <Gallery Match Number> <Found Flag> <Confidence>\n");

  for (i=0; i<no_files; i++){
    temp = faces[i].similar;
   /* fprintf(stderr, "in print_match faces[%d].name is %s\n", i, faces[i].name);
      fprintf(stderr, "in print_match temp->name is %s\n", temp->name);
      fprintf(stderr, "in print_match temp->next->name is %s\n", temp->next->name);*/
    count = 1;
    
    fprintf(fp, "\n\n");
    
    l = strlen(faces[i].name);
    for (k=0; k<l ; k++){
      if ( faces[i].name[k] >= '0' && faces[i].name[k] <= '9'){
	test[k] = faces[i].name[k];
      }
      else 
	break;
    }
    test[k]='\0';
    
    for (j=0; j<n; j++){
      l = strlen(temp->name);
      for (k=0; k<l; k++){
	if (temp->name[k] >= '0' && temp->name[k] <= '9')
	  known[k] = temp->name[k];
	else 
	  break;
      }
      known[k] = '\0';
      
      if (count == 1) {
	d  = temp->distance;
	if ( strcmp(test, known)== 0){
	    rec++;
	  }
      }
      
      /*e = 100*log(conf_factor)/log(temp->distance + conf_factor);*/
      /*	fprintf(stderr, "in print_match: temp->name is %s\n", temp->name);
		fprintf(stderr, "in print_match: temp->dist is %f\n\n", temp->distance);*/
      /* fprintf(fp, "%4d %3d %4d   %1d   %5.3f %f\n", atoi(test), count, atoi(known), (e < thres ? 0 : 1), e, temp->distance); */
      
      
      fprintf(fp, "%4d %3d %4d   %1d   %5.3f\n", atoi(test), count, atoi(known), (temp->distance < thres ? 0 : 1), temp->distance);
      temp = temp->next;
      count++;
      /* fprintf(stderr, "the value of e is %f\n", e);
	 fprintf(stderr, "the value of d is %f\n", d); */
    }
  }
  
  
  fprintf(stdout, "%f\n", (float) rec/no_files);
  fprintf(stderr, "\n\nThe recognition rate is  %f  (%d/%d)\n", 
	  (float) rec/no_files, rec, no_files);
  fprintf(fp, "\n\nThe recognition rate is  %f  (%d/%d)\n", 
	  (float) rec/no_files, rec, no_files);
  
  fclose(fp);
} /* Void print_match */


Matchptr add_match(Matchptr root, Matchptr new, int n)
{
  Matchptr temp, prev;
  int count = 0;

  if (root==NULL){
    root=new;
    return new;
  }
  else {
    if (root->distance < new->distance){
      new->next = root;
      return new;
    }
  }
  
  temp = root;
  prev = root;

  while ( ((temp != NULL) && (count != n))  && (temp->distance >= new->distance)){
    prev = temp;
    temp = temp->next;
    count++;
  }
  new->next = temp;
  prev->next = new;

  return root;
} /* void add_match(.......... */



int print_file(Tface tfaces, int n, FILE *fout, float thres, int count)
{
  int i, j, k, l, rec = 1;
  float d, e;
  Matchptr temp;
  
  char test[MAX_CHARS], known[MAX_CHARS];

  temp = tfaces.similar;
  
  l = strlen(tfaces.name);
  for (k=0; k<l ; k++){
    if ( tfaces.name[k] >= '0' && tfaces.name[k] <= '9'){
      test[k] = tfaces.name[k];
    }
    else 
      break;
  }
  test[k]='\0';
  
  
  for (j=0; j<n; j++){
    l = strlen(temp->name);
    for (k=0; k<l; k++){
      if (temp->name[k] >= '0' && temp->name[k] <= '9')
	known[k] = temp->name[k];
      else 
	break;
    }
    known[k] = '\0';
    
    if (rec == 1){
      d = temp->distance;
      if (strcmp(test, known) == 0){
	count++;
      }
    }
    
    fprintf(fout, "%4d %3d %4d   %1d   %5.3f\n", atoi(test), rec, atoi(known), (temp->distance < thres ? 0 : 1), temp->distance);
    temp = temp->next;
    rec++;
  }

  return(count);
}



void kill_list(Matchptr list)
{
  if (!empty_list(list)){
    kill_list(list->next);
    free(list->name);
    free(list);
  }
}

int empty_list(Matchptr list)
{
  if (list == NULL)
    return (1);
  return (0);
}





